try:
    from unittest import mock
except ImportError:
    import mock

from django.db import models
from django.db.models import Q
from django.test import TestCase

from flexi_tag.utils.compat import JSONField
from flexi_tag.utils.model_managers import TaggableManager
from flexi_tag.utils.models import FlexiTagMixin


class TaggableManagerTestModel(FlexiTagMixin):
    name = models.CharField(max_length=100)

    class Meta:
        app_label = "tests"


class TaggableManagerTestModelTag(models.Model):
    instance = models.OneToOneField(
        "tests.TaggableManagerTestModel",
        on_delete=models.CASCADE,
        primary_key=True,
    )
    tags = JSONField(default=list)

    class Meta:
        app_label = "tests"


class TaggableManagerComplexTestModel(FlexiTagMixin):
    name = models.CharField(max_length=100)

    class Meta:
        app_label = "tests"


class TaggableManagerComplexTestModelTag(models.Model):
    instance = models.OneToOneField(
        "tests.TaggableManagerComplexTestModel",
        on_delete=models.CASCADE,
        primary_key=True,
    )
    tags = JSONField(default=list)

    class Meta:
        app_label = "tests"


class TaggableManagerTestCase(TestCase):
    def setUp(self):
        self.manager = TaggableManager()
        self.manager.model = TaggableManagerTestModel

    @mock.patch("flexi_tag.utils.model_managers.apps.get_model")
    def test_validate_tag_model_class_success(self, mock_get_model):
        mock_get_model.return_value = TaggableManagerTestModelTag

        result = self.manager._TaggableManager__validate_tag_model_class(
            "tests", "TaggableManagerTestModelTag"
        )

        mock_get_model.assert_called_once_with("tests", "TaggableManagerTestModelTag")
        self.assertEqual(result, TaggableManagerTestModelTag)

    @mock.patch("flexi_tag.utils.model_managers.apps.get_model")
    def test_validate_tag_model_class_not_found(self, mock_get_model):
        mock_get_model.side_effect = LookupError()

        with self.assertRaises(ValueError) as context:
            self.manager._TaggableManager__validate_tag_model_class(
                "tests", "NonExistentModel"
            )

        self.assertIn(
            "Tag model for tests.NonExistentModel not found.", str(context.exception)
        )
        self.assertIn(
            "Did you run 'python manage.py generate_tag_models'", str(context.exception)
        )

    @mock.patch(
        "flexi_tag.utils.model_managers.TaggableManager._TaggableManager__validate_tag_model_class"
    )
    def test_get_tag_model_class_name(self, mock_validate):
        mock_validate.return_value = TaggableManagerTestModelTag

        result = self.manager._TaggableManager__get_tag_model_class_name(
            TaggableManagerTestModel
        )

        self.assertEqual(result, "TaggableManagerTestModelTag")
        mock_validate.assert_called_once_with(
            app_label="tests", tag_model_name="TaggableManagerTestModelTag"
        )

    @mock.patch(
        "flexi_tag.utils.model_managers.TaggableManager._TaggableManager__get_tag_model_class_name"
    )
    def test_filter_tag(self, mock_get_tag_model_class_name):
        mock_get_tag_model_class_name.return_value = "TaggableManagerTestModelTag"
        mock_queryset = mock.MagicMock()
        mock_filtered_queryset = mock.MagicMock()
        mock_queryset.filter.return_value = mock_filtered_queryset
        self.manager.get_queryset = mock.MagicMock(return_value=mock_queryset)

        result = self.manager.filter_tag("test_tag")

        mock_get_tag_model_class_name.assert_called_once_with(TaggableManagerTestModel)
        mock_queryset.filter.assert_called_once_with(
            **{"TaggableManagerTestModelTag__tags__contains": ["test_tag"]}
        )
        self.assertEqual(result, mock_filtered_queryset)

    @mock.patch(
        "flexi_tag.utils.model_managers.TaggableManager._TaggableManager__get_tag_model_class_name"
    )
    def test_exclude_tag(self, mock_get_tag_model_class_name):
        mock_get_tag_model_class_name.return_value = "TaggableManagerTestModelTag"
        mock_queryset = mock.MagicMock()
        mock_excluded_queryset = mock.MagicMock()
        mock_queryset.exclude.return_value = mock_excluded_queryset
        self.manager.get_queryset = mock.MagicMock(return_value=mock_queryset)

        result = self.manager.exclude_tag("test_tag")

        mock_get_tag_model_class_name.assert_called_once_with(TaggableManagerTestModel)
        mock_queryset.exclude.assert_called_once_with(
            **{"TaggableManagerTestModelTag__tags__contains": ["test_tag"]}
        )
        self.assertEqual(result, mock_excluded_queryset)

    @mock.patch(
        "flexi_tag.utils.model_managers.TaggableManager._TaggableManager__get_tag_model_class_name"
    )
    def test_with_tags(self, mock_get_tag_model_class_name):
        mock_get_tag_model_class_name.return_value = "TaggableManagerTestModelTag"
        mock_queryset = mock.MagicMock()
        mock_prefetched_queryset = mock.MagicMock()
        mock_queryset.prefetch_related.return_value = mock_prefetched_queryset
        self.manager.get_queryset = mock.MagicMock(return_value=mock_queryset)

        result = self.manager.with_tags()

        mock_get_tag_model_class_name.assert_called_once_with(TaggableManagerTestModel)
        mock_queryset.prefetch_related.assert_called_once_with(
            "TaggableManagerTestModelTag"
        )
        self.assertEqual(result, mock_prefetched_queryset)

    @mock.patch("django.apps.apps.get_model")
    def test_filter_tag_integration(self, mock_get_model):
        mock_tag_model = mock.MagicMock()
        mock_get_model.return_value = mock_tag_model

        manager = TaggableManager()
        manager.model = TaggableManagerTestModel

        mock_queryset = mock.MagicMock()
        mock_filtered_queryset = mock.MagicMock()
        mock_queryset.filter.return_value = mock_filtered_queryset
        manager.get_queryset = mock.MagicMock(return_value=mock_queryset)

        result = manager.filter_tag("test_tag")

        self.assertEqual(result, mock_filtered_queryset)
        self.assertEqual(mock_queryset.filter.call_count, 1)

    @mock.patch("django.apps.apps.get_model")
    def test_exclude_tag_integration(self, mock_get_model):
        mock_tag_model = mock.MagicMock()
        mock_get_model.return_value = mock_tag_model

        manager = TaggableManager()
        manager.model = TaggableManagerTestModel

        mock_queryset = mock.MagicMock()
        mock_excluded_queryset = mock.MagicMock()
        mock_queryset.exclude.return_value = mock_excluded_queryset
        manager.get_queryset = mock.MagicMock(return_value=mock_queryset)

        result = manager.exclude_tag("test_tag")

        self.assertEqual(result, mock_excluded_queryset)
        self.assertEqual(mock_queryset.exclude.call_count, 1)

    @mock.patch("django.apps.apps.get_model")
    def test_with_tags_integration(self, mock_get_model):
        mock_tag_model = mock.MagicMock()
        mock_get_model.return_value = mock_tag_model

        manager = TaggableManager()
        manager.model = TaggableManagerTestModel

        mock_queryset = mock.MagicMock()
        mock_prefetched_queryset = mock.MagicMock()
        mock_queryset.prefetch_related.return_value = mock_prefetched_queryset
        manager.get_queryset = mock.MagicMock(return_value=mock_queryset)

        result = manager.with_tags()

        self.assertEqual(result, mock_prefetched_queryset)
        self.assertEqual(mock_queryset.prefetch_related.call_count, 1)


class TaggableManagerComplexTestCase(TestCase):
    def setUp(self):
        self.manager = TaggableManager()
        self.manager.model = TaggableManagerComplexTestModel

    @mock.patch(
        "flexi_tag.utils.model_managers.TaggableManager._TaggableManager__get_tag_model_class_name"
    )
    def test_filter_tag_with_q_objects(self, mock_get_tag_model_class_name):
        mock_get_tag_model_class_name.return_value = (
            "TaggableManagerComplexTestModelTag"
        )
        mock_queryset = mock.MagicMock()
        mock_filtered_queryset = mock.MagicMock()
        mock_q_filtered_queryset = mock.MagicMock()

        mock_queryset.filter.return_value = mock_filtered_queryset
        mock_filtered_queryset.filter.return_value = mock_q_filtered_queryset

        self.manager.get_queryset = mock.MagicMock(return_value=mock_queryset)

        q_object = Q(status="active")

        filtered_queryset = self.manager.filter_tag("test_tag")
        result = filtered_queryset.filter(q_object)

        mock_get_tag_model_class_name.assert_called_once_with(
            TaggableManagerComplexTestModel
        )

        expected_filter = "TaggableManagerComplexTestModelTag__tags__contains"
        self.assertEqual(mock_queryset.filter.call_count, 1)
        call_kwargs = mock_queryset.filter.call_args[1]
        self.assertIn(expected_filter, call_kwargs)
        self.assertEqual(call_kwargs[expected_filter], ["test_tag"])

        self.assertEqual(mock_filtered_queryset.filter.call_count, 1)
        args = mock_filtered_queryset.filter.call_args[0]
        self.assertTrue(any(isinstance(arg, Q) for arg in args))

        self.assertEqual(result, mock_q_filtered_queryset)

    @mock.patch(
        "flexi_tag.utils.model_managers.TaggableManager._TaggableManager__get_tag_model_class_name"
    )
    def test_exclude_tag_with_q_objects(self, mock_get_tag_model_class_name):
        mock_get_tag_model_class_name.return_value = (
            "TaggableManagerComplexTestModelTag"
        )
        mock_queryset = mock.MagicMock()
        mock_excluded_queryset = mock.MagicMock()
        mock_q_excluded_queryset = mock.MagicMock()

        mock_queryset.exclude.return_value = mock_excluded_queryset
        mock_excluded_queryset.filter.return_value = mock_q_excluded_queryset

        self.manager.get_queryset = mock.MagicMock(return_value=mock_queryset)

        q_object = Q(status="inactive")

        excluded_queryset = self.manager.exclude_tag("test_tag")
        result = excluded_queryset.filter(q_object)

        mock_get_tag_model_class_name.assert_called_once_with(
            TaggableManagerComplexTestModel
        )

        expected_filter = "TaggableManagerComplexTestModelTag__tags__contains"
        self.assertEqual(mock_queryset.exclude.call_count, 1)
        call_kwargs = mock_queryset.exclude.call_args[1]
        self.assertIn(expected_filter, call_kwargs)
        self.assertEqual(call_kwargs[expected_filter], ["test_tag"])

        self.assertEqual(mock_excluded_queryset.filter.call_count, 1)
        args = mock_excluded_queryset.filter.call_args[0]
        self.assertTrue(any(isinstance(arg, Q) for arg in args))

        self.assertEqual(result, mock_q_excluded_queryset)

    @mock.patch(
        "flexi_tag.utils.model_managers.TaggableManager._TaggableManager__get_tag_model_class_name"
    )
    def test_with_tags_select_related(self, mock_get_tag_model_class_name):
        mock_get_tag_model_class_name.return_value = (
            "TaggableManagerComplexTestModelTag"
        )
        mock_queryset = mock.MagicMock()
        mock_prefetched_queryset = mock.MagicMock()
        mock_select_related_queryset = mock.MagicMock()

        mock_queryset.prefetch_related.return_value = mock_prefetched_queryset
        mock_prefetched_queryset.select_related.return_value = (
            mock_select_related_queryset
        )

        self.manager.get_queryset = mock.MagicMock(return_value=mock_queryset)

        result = self.manager.with_tags().select_related("some_related_field")

        mock_get_tag_model_class_name.assert_called_once_with(
            TaggableManagerComplexTestModel
        )
        mock_queryset.prefetch_related.assert_called_once_with(
            "TaggableManagerComplexTestModelTag"
        )
        mock_prefetched_queryset.select_related.assert_called_once_with(
            "some_related_field"
        )
        self.assertEqual(result, mock_select_related_queryset)

    @mock.patch("flexi_tag.utils.model_managers.apps.get_model")
    def test_validate_tag_model_class_recursive(self, mock_get_model):
        mock_get_model.return_value = TaggableManagerComplexTestModelTag

        result = self.manager._TaggableManager__validate_tag_model_class(
            "tests", "TaggableManagerComplexTestModelTag"
        )

        mock_get_model.assert_called_once_with(
            "tests", "TaggableManagerComplexTestModelTag"
        )
        self.assertEqual(result, TaggableManagerComplexTestModelTag)

    def test_get_tag_model_class_name_with_nested_classes(self):
        class OuterTestModel:
            class InnerTestModel(FlexiTagMixin):
                class Meta:
                    app_label = "tests"

        self.manager._TaggableManager__validate_tag_model_class = mock.MagicMock()

        result = self.manager._TaggableManager__get_tag_model_class_name(
            OuterTestModel.InnerTestModel
        )

        self.assertEqual(result, "InnerTestModelTag")
        self.manager._TaggableManager__validate_tag_model_class.assert_called_once_with(
            app_label="tests", tag_model_name="InnerTestModelTag"
        )



from django.utils.encoding import smart_str

from flexi_tag import codes


class ProjectBaseException(Exception):
    code = codes.undefined

    def __init__(self, *args, **kwargs):
        if not isinstance(self.code, dict):
            raise Exception("parameter type must be a dict")
        code = self.code.get("code", "undefined")
        message = getattr(self.codes, "%s" % code)
        self.message = message.get("en")
        self.obj = kwargs.get("obj", None)
        self.target = kwargs.get("target", None)
        self.params = kwargs.get("params")
        if self.params and isinstance(self.params, dict):
            self.message = smart_str(self.message).format(**self.params)
        elif self.params and isinstance(self.params, (list, set, tuple)):
            self.message = smart_str(self.message).format(*self.params)

        Exception.__init__(self, smart_str("{0}:{1}").format(code, self.message))

    def __new__(cls, *args, **kwargs):
        obj = super(ProjectBaseException, cls).__new__(cls)
        obj.__init__(*args, **kwargs)
        try:
            getattr(cls.codes, "%s" % obj.code.get("code"))
        except AttributeError:
            pass
        return obj

    @property
    def codes(self):
        return codes


class TagValidationException(ProjectBaseException):
    code = codes.tag_100_1


class TagNotFoundException(ProjectBaseException):
    code = codes.tag_100_2


class TagNotDefinedException(ProjectBaseException):
    code = codes.tag_100_3


class ObjectIDsNotDefinedException(ProjectBaseException):
    code = codes.tag_100_4



import inspect
import os
from importlib import import_module

from django.apps import apps
from django.core.management.base import BaseCommand
from django.template import Context, Template

from flexi_tag.utils.models import FlexiTagMixin


class Command(BaseCommand):
    help = "Generate tag models for all models that inherit from FlexiTagMixin"

    def add_arguments(self, parser):
        parser.add_argument(
            "--dry-run",
            action="store_true",
            help="Show what would be generated without creating files",
        )

    def handle(self, *args, **options):
        dry_run = options.get("dry_run", False)
        # Template for the generated model
        model_template = Template(
            """
# This file is auto-generated. Do not edit manually.
# Generated by the generate_tag_models command.

from django.db import models
from flexi_tag.utils.compat import JSONField
from django.contrib.postgres.indexes import GinIndex


class {{ model_name }}Tag(models.Model):
    instance = models.OneToOneField(
        "{{ app_label }}.{{ model_name }}",
        on_delete=models.CASCADE,
        primary_key=True,
    )
    tags = JSONField(default=list)

    class Meta:
        app_label = "{{ app_label }}"
        db_table = "{{ app_label }}_{{ model_lower_name }}_tag"
        indexes = [GinIndex(fields=["tags"])]

    def __str__(self):
        return "Tags for {}".format(self.instance)
"""
        )

        for app_config in apps.get_app_configs():
            for model in app_config.get_models():
                if issubclass(model, FlexiTagMixin) and model is not FlexiTagMixin:
                    model_name = model.__name__
                    app_label = model._meta.app_label  # noqa
                    model_module = import_module(model.__module__)
                    model_dir = os.path.dirname(inspect.getfile(model_module))

                    context = Context(
                        {
                            "model_name": model_name,
                            "app_label": app_label,
                            "model_lower_name": model_name.lower(),
                        }
                    )

                    model_content = model_template.render(context)

                    file_path = os.path.join(model_dir, "flexi_generated_model.py")

                    if dry_run:
                        self.stdout.write(
                            "Would create file: {file_path}".format(file_path=file_path)
                        )
                        self.stdout.write(model_content)
                        self.stdout.write("-" * 40)
                    else:
                        with open(file_path, "w") as f:
                            f.write(model_content)
                        self.stdout.write(
                            self.style.SUCCESS(
                                "Created tag model for {app_label}.{model_name} at {file_path}".format(
                                    app_label=app_label,
                                    model_name=model_name,
                                    file_path=file_path,
                                )
                            )
                        )

        if dry_run:
            self.stdout.write("Dry run completed. No files were created.")
        else:
            self.stdout.write(
                self.style.SUCCESS("Successfully generated all tag models!")
            )



try:
    from unittest import mock
except ImportError:
    import mock

from django.db import models
from django.db.models import QuerySet
from django.test import TestCase

from flexi_tag.exceptions import TagNotFoundException, TagValidationException
from flexi_tag.utils.compat import JSONField
from flexi_tag.utils.models import FlexiTagMixin
from flexi_tag.utils.service import TaggableService


class ServiceTestModel(FlexiTagMixin):
    name = models.CharField(max_length=100)

    class Meta:
        app_label = "tests"


class ServiceTestModelTag(models.Model):
    instance = models.OneToOneField(
        "tests.ServiceTestModel",
        on_delete=models.CASCADE,
        primary_key=True,
    )
    tags = JSONField(default=list)

    class Meta:
        app_label = "tests"


class TaggableServiceTestCase(TestCase):
    def setUp(self):
        self.service = TaggableService()
        self.test_model = mock.MagicMock(spec=ServiceTestModel)
        self.test_model._meta.app_label = "tests"
        self.test_model.__class__.__name__ = "ServiceTestModel"

    @mock.patch("flexi_tag.utils.service.apps.get_model")
    def test_get_tag_model_class(self, mock_get_model):
        mock_get_model.return_value = ServiceTestModelTag

        result = TaggableService._TaggableService__get_tag_model_class(self.test_model)

        mock_get_model.assert_called_once_with("tests", "ServiceTestModelTag")
        self.assertEqual(result, ServiceTestModelTag)

    @mock.patch("flexi_tag.utils.service.apps.get_model")
    def test_get_tag_model_class_not_found(self, mock_get_model):
        mock_get_model.side_effect = LookupError()

        with self.assertRaises(ValueError) as context:
            TaggableService._TaggableService__get_tag_model_class(self.test_model)

        self.assertIn(
            "Tag model for tests.ServiceTestModel not found", str(context.exception)
        )
        self.assertIn(
            "Did you run 'python manage.py generate_tag_models'", str(context.exception)
        )

    def test_validate_tag_key_success(self):
        result = TaggableService._TaggableService__validate_tag_key("valid_key")

        self.assertTrue(result)

    def test_validate_tag_key_failure(self):
        for invalid_key in [123, True, None, [], {}]:
            with self.assertRaises(ValueError) as context:
                TaggableService._TaggableService__validate_tag_key(invalid_key)

            self.assertEqual(str(context.exception), "Tag key must be a string.")

    @mock.patch(
        "flexi_tag.utils.service.TaggableService._TaggableService__get_tag_model_class"
    )
    def test_add_tag(self, mock_get_tag_model_class):
        mock_tag_model = mock.MagicMock()
        mock_get_tag_model_class.return_value = mock_tag_model

        mock_tag_instance = mock.MagicMock()
        mock_tag_instance.tags = []

        mock_tag_model.objects.get_or_create.return_value = (mock_tag_instance, True)

        result = self.service.add_tag(self.test_model, "test_tag")

        mock_get_tag_model_class.assert_called_once_with(self.test_model)
        mock_tag_model.objects.get_or_create.assert_called_once_with(
            instance=self.test_model
        )
        self.assertEqual(mock_tag_instance.tags, ["test_tag"])
        mock_tag_instance.save.assert_called_once_with(update_fields=["tags"])
        self.assertEqual(result, self.test_model)

    @mock.patch(
        "flexi_tag.utils.service.TaggableService._TaggableService__get_tag_model_class"
    )
    def test_add_tag_already_exists(self, mock_get_tag_model_class):
        mock_tag_model = mock.MagicMock()
        mock_get_tag_model_class.return_value = mock_tag_model

        mock_tag_instance = mock.MagicMock()
        mock_tag_instance.tags = ["existing_tag"]

        mock_tag_model.objects.get_or_create.return_value = (mock_tag_instance, False)

        with self.assertRaises(TagValidationException):
            self.service.add_tag(self.test_model, "existing_tag")

    @mock.patch(
        "flexi_tag.utils.service.TaggableService._TaggableService__get_tag_model_class"
    )
    def test_bulk_add_tags(self, mock_get_tag_model_class):
        mock_tag_model = mock.MagicMock()
        mock_get_tag_model_class.return_value = mock_tag_model

        mock_tag_instance = mock.MagicMock()
        mock_tag_instance.tags = []

        mock_tag_model.objects.get_or_create.return_value = (mock_tag_instance, True)

        result = self.service.bulk_add_tags(self.test_model, ["tag1", "tag2", "tag3"])

        mock_get_tag_model_class.assert_called_once_with(self.test_model)
        mock_tag_model.objects.get_or_create.assert_called_once_with(
            instance=self.test_model
        )
        self.assertEqual(mock_tag_instance.tags, ["tag1", "tag2", "tag3"])
        mock_tag_instance.save.assert_called_once_with(update_fields=["tags"])
        self.assertEqual(result, self.test_model)

    @mock.patch(
        "flexi_tag.utils.service.TaggableService._TaggableService__get_tag_model_class"
    )
    def test_bulk_add_tags_already_exists(self, mock_get_tag_model_class):
        mock_tag_model = mock.MagicMock()
        mock_get_tag_model_class.return_value = mock_tag_model

        mock_tag_instance = mock.MagicMock()
        mock_tag_instance.tags = ["existing_tag"]

        mock_tag_model.objects.get_or_create.return_value = (mock_tag_instance, False)

        with self.assertRaises(TagValidationException):
            self.service.bulk_add_tags(self.test_model, ["new_tag", "existing_tag"])

    @mock.patch(
        "flexi_tag.utils.service.TaggableService._TaggableService__get_tag_model_class"
    )
    def test_bulk_add_tags_with_many_instances(self, mock_get_tag_model_class):
        mock_tag_model = mock.MagicMock()
        mock_get_tag_model_class.return_value = mock_tag_model

        instance1 = mock.MagicMock()
        instance2 = mock.MagicMock()
        mock_queryset = mock.MagicMock(spec=QuerySet)
        mock_queryset.__getitem__.return_value = instance1
        mock_queryset.iterator.return_value = [instance1, instance2]

        mock_tag_instance1 = mock.MagicMock()
        mock_tag_instance1.tags = []

        mock_tag_instance2 = mock.MagicMock()
        mock_tag_instance2.tags = []

        mock_tag_model.objects.get_or_create.side_effect = [
            (mock_tag_instance1, True),
            (mock_tag_instance2, True),
        ]

        result = self.service.bulk_add_tags_with_many_instances(
            mock_queryset, ["tag1", "tag2"]
        )

        self.assertEqual(mock_get_tag_model_class.call_count, 1)
        self.assertEqual(mock_tag_model.objects.get_or_create.call_count, 2)
        self.assertEqual(mock_tag_instance1.save.call_count, 1)
        self.assertEqual(mock_tag_instance2.save.call_count, 1)
        self.assertEqual(result, mock_queryset)

    @mock.patch(
        "flexi_tag.utils.service.TaggableService._TaggableService__get_tag_model_class"
    )
    def test_bulk_add_tags_with_many_instances_tag_validation(
        self, mock_get_tag_model_class
    ):
        mock_tag_model = mock.MagicMock()
        mock_get_tag_model_class.return_value = mock_tag_model

        instance1 = mock.MagicMock()
        instance2 = mock.MagicMock()

        mock_queryset = mock.MagicMock(spec=QuerySet)
        mock_queryset.__getitem__.return_value = instance1
        mock_queryset.iterator.return_value = [instance1, instance2]

        mock_tag_instance1 = mock.MagicMock()
        mock_tag_instance1.tags = ["existing_tag"]

        mock_tag_model.objects.get_or_create.return_value = (mock_tag_instance1, False)

        with self.assertRaises(ValueError):
            self.service.bulk_add_tags_with_many_instances(
                mock_queryset, ["valid_tag", 123]
            )

        self.assertEqual(mock_get_tag_model_class.call_count, 1)
        self.assertEqual(mock_tag_model.objects.get_or_create.call_count, 1)

    @mock.patch(
        "flexi_tag.utils.service.TaggableService._TaggableService__get_tag_model_class"
    )
    def test_bulk_add_tags_with_many_instances_tag_in_instance(
        self, mock_get_tag_model_class
    ):
        mock_tag_model = mock.MagicMock()
        mock_get_tag_model_class.return_value = mock_tag_model

        instance1 = mock.MagicMock()

        mock_queryset = mock.MagicMock(spec=QuerySet)
        mock_queryset.__getitem__.return_value = instance1
        mock_queryset.iterator.return_value = [instance1]

        mock_tag_instance = mock.MagicMock()
        mock_tag_instance.tags = ["existing_tag"]

        mock_tag_model.objects.get_or_create.return_value = (mock_tag_instance, False)

        with self.assertRaises(TagValidationException):
            self.service.bulk_add_tags_with_many_instances(
                mock_queryset, ["existing_tag"]
            )

        self.assertEqual(mock_get_tag_model_class.call_count, 1)
        self.assertEqual(mock_tag_model.objects.get_or_create.call_count, 1)

    @mock.patch(
        "flexi_tag.utils.service.TaggableService._TaggableService__get_tag_model_class"
    )
    def test_remove_tag(self, mock_get_tag_model_class):
        mock_tag_model = mock.MagicMock()
        mock_get_tag_model_class.return_value = mock_tag_model

        mock_tag_instance = mock.MagicMock()
        tags_list = ["tag_to_remove", "other_tag"]
        mock_tag_instance.tags = tags_list

        mock_tag_model.objects.filter.return_value.first.return_value = (
            mock_tag_instance
        )

        result = self.service.remove_tag(self.test_model, "tag_to_remove")

        self.assertEqual(mock_get_tag_model_class.call_count, 1)
        mock_tag_model.objects.filter.assert_called_once_with(instance=self.test_model)
        self.assertEqual(len(tags_list), 1)
        self.assertEqual(tags_list[0], "other_tag")
        mock_tag_instance.save.assert_called_once_with(update_fields=["tags"])
        self.assertEqual(result, mock_tag_instance)

    @mock.patch(
        "flexi_tag.utils.service.TaggableService._TaggableService__get_tag_model_class"
    )
    def test_remove_tag_instance_not_found(self, mock_get_tag_model_class):
        mock_tag_model = mock.MagicMock()
        mock_get_tag_model_class.return_value = mock_tag_model

        mock_filter = mock.MagicMock()
        mock_filter.first.return_value = None
        mock_tag_model.objects.filter.return_value = mock_filter

        with self.assertRaises(TagNotFoundException):
            self.service.remove_tag(self.test_model, "any_tag")

        mock_get_tag_model_class.assert_called_once_with(self.test_model)
        mock_tag_model.objects.filter.assert_called_once_with(instance=self.test_model)

    @mock.patch(
        "flexi_tag.utils.service.TaggableService._TaggableService__get_tag_model_class"
    )
    def test_remove_tag_not_found_specific_with_patch(self, mock_get_tag_model_class):
        mock_tag_model = mock.MagicMock()
        mock_get_tag_model_class.return_value = mock_tag_model

        def filter_side_effect(*args, **kwargs):
            mock_filter = mock.MagicMock()
            mock_filter.first.return_value = None
            return mock_filter

        mock_tag_model.objects.filter = mock.MagicMock(side_effect=filter_side_effect)

        with self.assertRaises(TagNotFoundException):
            self.service.remove_tag(self.test_model, "test_tag")

        mock_get_tag_model_class.assert_called_once_with(self.test_model)
        mock_tag_model.objects.filter.assert_called_once_with(instance=self.test_model)

    @mock.patch(
        "flexi_tag.utils.service.TaggableService._TaggableService__get_tag_model_class"
    )
    def test_remove_tag_tag_not_in_instance(self, mock_get_tag_model_class):
        mock_tag_model = mock.MagicMock()
        mock_get_tag_model_class.return_value = mock_tag_model

        mock_tag_instance = mock.MagicMock()
        mock_tag_instance.tags = ["tag1", "tag2"]

        mock_filter = mock.MagicMock()
        mock_filter.first.return_value = mock_tag_instance
        mock_tag_model.objects.filter.return_value = mock_filter

        with self.assertRaises(TagValidationException):
            self.service.remove_tag(self.test_model, "non_existing_tag")

        mock_get_tag_model_class.assert_called_once_with(self.test_model)
        mock_tag_model.objects.filter.assert_called_once_with(instance=self.test_model)
        mock_tag_instance.save.assert_not_called()

    @mock.patch(
        "flexi_tag.utils.service.TaggableService._TaggableService__get_tag_model_class"
    )
    def test_remove_tag_validation_exception_specific(self, mock_get_tag_model_class):
        mock_tag_model = mock.MagicMock()
        mock_get_tag_model_class.return_value = mock_tag_model

        mock_tag_instance = mock.MagicMock()
        mock_tag_instance.tags = ["tag1", "tag2"]

        mock_filter = mock.MagicMock()
        mock_filter.first.return_value = mock_tag_instance
        mock_tag_model.objects.filter.return_value = mock_filter

        specific_tag = "missing_tag"
        with self.assertRaises(TagValidationException):
            self.service.remove_tag(self.test_model, specific_tag)

        mock_get_tag_model_class.assert_called_once_with(self.test_model)
        mock_tag_model.objects.filter.assert_called_once_with(instance=self.test_model)

    @mock.patch(
        "flexi_tag.utils.service.TaggableService._TaggableService__get_tag_model_class"
    )
    def test_remove_tag_tag_not_in_instance_with_patch(self, mock_get_tag_model_class):
        mock_tag_model = mock.MagicMock()
        mock_get_tag_model_class.return_value = mock_tag_model

        tags_list = ["tag1", "tag2"]
        mock_tag_instance = mock.MagicMock()

        mock_tag_instance.tags = tags_list

        def filter_side_effect(*args, **kwargs):
            mock_filter = mock.MagicMock()
            mock_filter.first.return_value = mock_tag_instance
            return mock_filter

        mock_tag_model.objects.filter = mock.MagicMock(side_effect=filter_side_effect)

        with self.assertRaises(TagValidationException):
            tag_not_in_list = "not_in_list"
            self.service.remove_tag(self.test_model, tag_not_in_list)

        mock_get_tag_model_class.assert_called_once_with(self.test_model)
        mock_tag_model.objects.filter.assert_called_once_with(instance=self.test_model)

    @mock.patch(
        "flexi_tag.utils.service.TaggableService._TaggableService__get_tag_model_class"
    )
    def test_bulk_remove_tags(self, mock_get_tag_model_class):
        mock_tag_model = mock.MagicMock()
        mock_get_tag_model_class.return_value = mock_tag_model

        tags_list = ["tag1", "tag2", "tag3"]
        mock_tag_instance = mock.MagicMock()
        mock_tag_instance.tags = tags_list

        mock_tag_model.objects.filter.return_value.first.return_value = (
            mock_tag_instance
        )

        keys = ["tag1", "tag3"]
        result = self.service.bulk_remove_tags(self.test_model, keys)

        mock_get_tag_model_class.assert_called_once_with(self.test_model)
        mock_tag_model.objects.filter.assert_called_once_with(instance=self.test_model)

        self.assertEqual(len(tags_list), 1)
        self.assertEqual(tags_list[0], "tag2")
        mock_tag_instance.save.assert_called_once_with(update_fields=["tags"])
        self.assertEqual(result, mock_tag_instance)

    @mock.patch(
        "flexi_tag.utils.service.TaggableService._TaggableService__get_tag_model_class"
    )
    def test_bulk_remove_tags_not_found_instance(self, mock_get_tag_model_class):
        mock_tag_model = mock.MagicMock()
        mock_get_tag_model_class.return_value = mock_tag_model

        mock_tag_model.objects.filter.return_value.first.return_value = None

        with self.assertRaises(TagNotFoundException):
            self.service.bulk_remove_tags(self.test_model, ["tag1", "tag2"])

    @mock.patch(
        "flexi_tag.utils.service.TaggableService._TaggableService__get_tag_model_class"
    )
    def test_bulk_remove_tags_not_exists(self, mock_get_tag_model_class):
        mock_tag_model = mock.MagicMock()
        mock_get_tag_model_class.return_value = mock_tag_model

        mock_tag_instance = mock.MagicMock()
        mock_tag_instance.tags = mock.MagicMock()
        mock_tag_instance.tags.__contains__.side_effect = lambda key: key == "tag1"

        mock_tag_model.objects.filter.return_value.first.return_value = (
            mock_tag_instance
        )

        with self.assertRaises(TagValidationException):
            self.service.bulk_remove_tags(self.test_model, ["tag1", "non_existent_tag"])

    @mock.patch(
        "flexi_tag.utils.service.TaggableService._TaggableService__get_tag_model_class"
    )
    def test_bulk_remove_tags_with_many_instances(self, mock_get_tag_model_class):
        mock_tag_model = mock.MagicMock()
        mock_get_tag_model_class.return_value = mock_tag_model

        instance1 = mock.MagicMock()
        instance2 = mock.MagicMock()
        mock_queryset = mock.MagicMock(spec=QuerySet)
        mock_queryset.__getitem__.return_value = instance1
        mock_queryset.iterator.return_value = [instance1, instance2]

        tags_list1 = ["tag1", "tag2"]
        mock_tag_instance1 = mock.MagicMock()
        mock_tag_instance1.tags = tags_list1

        tags_list2 = ["tag1", "tag3"]
        mock_tag_instance2 = mock.MagicMock()
        mock_tag_instance2.tags = tags_list2

        mock_tag_model.objects.filter.return_value.first.side_effect = [
            mock_tag_instance1,
            mock_tag_instance2,
        ]

        result = self.service.bulk_remove_tags_with_many_instances(
            mock_queryset, ["tag1"]
        )

        self.assertEqual(mock_get_tag_model_class.call_count, 1)
        self.assertEqual(mock_tag_model.objects.filter.call_count, 2)

        self.assertEqual(len(tags_list1), 1)
        self.assertEqual(tags_list1[0], "tag2")
        self.assertEqual(len(tags_list2), 1)
        self.assertEqual(tags_list2[0], "tag3")

        self.assertEqual(mock_tag_instance1.save.call_count, 1)
        self.assertEqual(mock_tag_instance2.save.call_count, 1)
        self.assertEqual(result, mock_queryset)

    @mock.patch(
        "flexi_tag.utils.service.TaggableService._TaggableService__get_tag_model_class"
    )
    def test_get_tags(self, mock_get_tag_model_class):
        mock_tag_model = mock.MagicMock()
        mock_get_tag_model_class.return_value = mock_tag_model

        mock_tag_instance = mock.MagicMock()
        mock_tag_instance.tags = ["tag1", "tag2"]

        mock_tag_model.objects.get.return_value = mock_tag_instance

        result = self.service.get_tags(self.test_model)

        mock_get_tag_model_class.assert_called_once_with(self.test_model)
        mock_tag_model.objects.get.assert_called_once_with(instance=self.test_model)
        self.assertEqual(result, ["tag1", "tag2"])

    @mock.patch(
        "flexi_tag.utils.service.TaggableService._TaggableService__get_tag_model_class"
    )
    def test_get_tags_none_result(self, mock_get_tag_model_class):
        mock_tag_model = mock.MagicMock()
        mock_get_tag_model_class.return_value = mock_tag_model

        mock_tag_model.objects.get.return_value = None

        result = self.service.get_tags(self.test_model)

        mock_get_tag_model_class.assert_called_once_with(self.test_model)
        mock_tag_model.objects.get.assert_called_once_with(instance=self.test_model)
        self.assertEqual(result, [])

    def test_filter_by_tag(self):
        mock_queryset = mock.MagicMock()
        mock_filtered_queryset = mock.MagicMock()
        mock_queryset.filter_tag.return_value = mock_filtered_queryset

        result = TaggableService.filter_by_tag(mock_queryset, "test_tag")

        mock_queryset.filter_tag.assert_called_once_with("test_tag")
        self.assertEqual(result, mock_filtered_queryset)

    def test_exclude_by_tag(self):
        mock_queryset = mock.MagicMock()
        mock_excluded_queryset = mock.MagicMock()
        mock_queryset.exclude_tag.return_value = mock_excluded_queryset

        result = TaggableService.exclude_by_tag(mock_queryset, "test_tag")

        mock_queryset.exclude_tag.assert_called_once_with("test_tag")
        self.assertEqual(result, mock_excluded_queryset)

    def test_get_with_tags(self):
        mock_queryset = mock.MagicMock()
        mock_prefetched_queryset = mock.MagicMock()
        mock_queryset.with_tags.return_value = mock_prefetched_queryset

        result = TaggableService.get_with_tags(mock_queryset)

        self.assertEqual(mock_queryset.with_tags.call_count, 1)
        self.assertEqual(result, mock_prefetched_queryset)

    @mock.patch("django.apps.apps.get_model")
    def test_add_and_remove_tag_integration(self, mock_get_model):
        mock_tag_model = mock.MagicMock()
        mock_get_model.return_value = mock_tag_model

        instance = ServiceTestModel(name="Test Integration")

        tags_list = []
        mock_tag_instance = mock.MagicMock()
        mock_tag_instance.tags = tags_list

        mock_tag_model.objects.get_or_create.return_value = (mock_tag_instance, True)

        mock_tag_model.objects.filter.return_value.first.return_value = (
            mock_tag_instance
        )

        self.service.add_tag(instance, "temp_tag")

        self.assertEqual(tags_list, ["temp_tag"])

        self.service.remove_tag(instance, "temp_tag")

        self.assertEqual(tags_list, [])

        mock_get_model.assert_called_with("tests", "ServiceTestModelTag")
        mock_tag_model.objects.get_or_create.assert_called_once_with(instance=instance)
        mock_tag_model.objects.filter.assert_called_once_with(instance=instance)



try:
    from unittest import mock
except ImportError:
    import mock

from django.db import models
from django.test import TestCase
from rest_framework.response import Response
from rest_framework.viewsets import ModelViewSet

from flexi_tag.exceptions import (
    ObjectIDsNotDefinedException,
    TagNotDefinedException,
    TagNotFoundException,
)
from flexi_tag.utils.models import FlexiTagMixin
from flexi_tag.utils.views import TaggableViewSetMixin


class TaggableTestModel(FlexiTagMixin):
    name = models.CharField(max_length=100)

    class Meta:
        app_label = "tests"


class TestViewSet(TaggableViewSetMixin, ModelViewSet):
    queryset = TaggableTestModel.objects.all()


class TaggableViewSetMixinTestCase(TestCase):
    def setUp(self):
        self.viewset = TestViewSet()
        self.viewset.format_kwarg = None
        self.viewset.taggable_service = mock.MagicMock()
        self.mock_object = mock.MagicMock(spec=TaggableTestModel)
        self.viewset.get_object = mock.MagicMock(return_value=self.mock_object)

        self.mock_queryset = mock.MagicMock()
        self.mock_filtered_queryset = mock.MagicMock()
        self.mock_queryset.filter.return_value = self.mock_filtered_queryset
        self.viewset.get_queryset = mock.MagicMock(return_value=self.mock_queryset)

        self.viewset.paginate_queryset = mock.MagicMock(return_value=None)

        mock_serializer = mock.MagicMock()
        mock_serializer.data = {"data": "serialized_data"}
        self.viewset.get_serializer = mock.MagicMock(return_value=mock_serializer)

    def test_add_tag(self):
        mock_request = mock.MagicMock()
        mock_request.data = {"key": "test_tag"}

        response = self.viewset.add_tag(mock_request, pk=1)

        self.viewset.taggable_service.add_tag.assert_called_once_with(
            instance=self.mock_object,
            key="test_tag",
        )

        self.assertEqual(response.status_code, 200)

    def test_add_tag_missing_key(self):
        mock_request = mock.MagicMock()
        mock_request.data = {}

        with self.assertRaises(TagNotDefinedException):
            self.viewset.add_tag(mock_request, pk=1)

    def test_bulk_add_tag(self):
        mock_request = mock.MagicMock()
        mock_request.data = {"keys": ["tag1", "tag2"]}

        response = self.viewset.bulk_add_tag(mock_request, pk=1)

        self.viewset.taggable_service.bulk_add_tags.assert_called_once_with(
            instance=self.mock_object,
            keys=["tag1", "tag2"],
        )

        self.assertEqual(response.status_code, 200)

    def test_bulk_add_tag_missing_keys(self):
        mock_request = mock.MagicMock()
        mock_request.data = {}

        with self.assertRaises(TagNotDefinedException):
            self.viewset.bulk_add_tag(mock_request, pk=1)

    def test_bulk_add_tags(self):
        mock_request = mock.MagicMock()
        mock_request.data = {"objects": [1, 2], "keys": ["tag1", "tag2"]}

        response = self.viewset.bulk_add_tags(mock_request)

        self.mock_queryset.filter.assert_called_once_with(id__in=[1, 2])

        self.viewset.taggable_service.bulk_add_tags_with_many_instances.assert_called_once_with(
            instances=self.mock_filtered_queryset,
            keys=["tag1", "tag2"],
        )

        self.assertEqual(response.status_code, 200)

    def test_bulk_add_tags_missing_keys(self):
        mock_request = mock.MagicMock()
        mock_request.data = {"objects": [1, 2]}

        with self.assertRaises(TagNotDefinedException):
            self.viewset.bulk_add_tags(mock_request)

    def test_bulk_add_tags_missing_objects(self):
        mock_request = mock.MagicMock()
        mock_request.data = {"keys": ["tag1", "tag2"]}

        with self.assertRaises(ObjectIDsNotDefinedException):
            self.viewset.bulk_add_tags(mock_request)

    def test_remove_tag(self):
        mock_request = mock.MagicMock()
        mock_request.data = {"key": "test_tag"}

        response = self.viewset.remove_tag(mock_request, pk=1)

        self.viewset.taggable_service.remove_tag.assert_called_once_with(
            instance=self.mock_object,
            key="test_tag",
        )

        self.assertEqual(response.status_code, 200)

    def test_remove_tag_missing_key(self):
        mock_request = mock.MagicMock()
        mock_request.data = {}

        with self.assertRaises(TagNotFoundException):
            self.viewset.remove_tag(mock_request, pk=1)

    def test_bulk_remove_tags(self):
        mock_request = mock.MagicMock()
        mock_request.data = {"keys": ["tag1", "tag2"]}

        response = self.viewset.bulk_remove_tags(mock_request, pk=1)

        self.viewset.taggable_service.bulk_remove_tags.assert_called_once_with(
            instance=self.mock_object,
            keys=["tag1", "tag2"],
        )

        self.assertEqual(response.status_code, 200)

    def test_bulk_remove_tags_missing_keys(self):
        mock_request = mock.MagicMock()
        mock_request.data = {}

        with self.assertRaises(TagNotDefinedException):
            self.viewset.bulk_remove_tags(mock_request, pk=1)

    def test_bulk_remove_tags_with_many_instances(self):
        mock_request = mock.MagicMock()
        mock_request.data = {"objects": [1, 2], "keys": ["tag1", "tag2"]}

        response = self.viewset.bulk_remove_tags_with_many_instances(mock_request)

        self.mock_queryset.filter.assert_called_once_with(id__in=[1, 2])

        self.viewset.taggable_service.bulk_remove_tags_with_many_instances.assert_called_once_with(
            instances=self.mock_filtered_queryset,
            keys=["tag1", "tag2"],
        )

        self.assertEqual(response.status_code, 200)

    def test_bulk_remove_tags_with_many_instances_missing_keys(self):
        mock_request = mock.MagicMock()
        mock_request.data = {"objects": [1, 2]}

        with self.assertRaises(TagNotDefinedException):
            self.viewset.bulk_remove_tags_with_many_instances(mock_request)

    def test_bulk_remove_tags_with_many_instances_missing_objects(self):
        mock_request = mock.MagicMock()
        mock_request.data = {"keys": ["tag1", "tag2"]}

        with self.assertRaises(ObjectIDsNotDefinedException):
            self.viewset.bulk_remove_tags_with_many_instances(mock_request)

    def test_get_tags(self):
        mock_request = mock.MagicMock()

        tag_instance = {"tags": ["tag1", "tag2"]}
        self.viewset.taggable_service.get_tags.return_value = tag_instance

        response = self.viewset.get_tags(mock_request, pk=1)

        self.viewset.taggable_service.get_tags.assert_called_once_with(
            instance=self.mock_object,
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data, tag_instance)

    def test_get_tags_not_found(self):
        mock_request = mock.MagicMock()

        self.viewset.taggable_service.get_tags.return_value = None

        response = self.viewset.get_tags(mock_request, pk=1)

        self.assertEqual(self.viewset.taggable_service.get_tags.call_count, 1)

        self.assertEqual(response.status_code, 404)

    def test_filter_by_tag(self):
        mock_request = mock.MagicMock()
        mock_request.query_params = {"key": "test_tag"}
        mock_request.data = {}

        self.viewset.taggable_service.filter_by_tag.return_value = (
            self.mock_filtered_queryset
        )

        response = self.viewset.filter_by_tag(mock_request)

        self.viewset.taggable_service.filter_by_tag.assert_called_once_with(
            queryset=self.mock_queryset,
            key="test_tag",
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data, {"data": "serialized_data"})

    def test_filter_by_tag_missing_key(self):
        mock_request = mock.MagicMock()
        mock_request.query_params = {}
        mock_request.data = {}

        with self.assertRaises(TagNotDefinedException):
            self.viewset.filter_by_tag(mock_request)

    def test_filter_by_tag_with_pagination(self):
        mock_request = mock.MagicMock()
        mock_request.query_params = {"key": "test_tag"}
        mock_request.data = {}

        page = [self.mock_object]
        self.viewset.paginate_queryset.return_value = page

        paginated_response = Response({"results": {"data": "serialized_data"}})
        self.viewset.get_paginated_response = mock.MagicMock(
            return_value=paginated_response
        )

        self.viewset.taggable_service.filter_by_tag.return_value = (
            self.mock_filtered_queryset
        )

        response = self.viewset.filter_by_tag(mock_request)

        self.viewset.paginate_queryset.assert_called_once_with(
            self.mock_filtered_queryset
        )

        self.viewset.get_serializer.assert_called_once_with(page, many=True)

        self.viewset.get_paginated_response.assert_called_once_with(
            {"data": "serialized_data"}
        )

        self.assertEqual(response, paginated_response)

    def test_exclude_by_tag(self):
        mock_request = mock.MagicMock()
        mock_request.query_params = {"key": "test_tag"}
        mock_request.data = {}

        self.viewset.taggable_service.exclude_by_tag.return_value = (
            self.mock_filtered_queryset
        )

        response = self.viewset.exclude_by_tag(mock_request)

        self.viewset.taggable_service.exclude_by_tag.assert_called_once_with(
            queryset=self.mock_queryset,
            key="test_tag",
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data, {"data": "serialized_data"})

    def test_exclude_by_tag_missing_key(self):
        mock_request = mock.MagicMock()
        mock_request.query_params = {}
        mock_request.data = {}

        with self.assertRaises(TagNotDefinedException):
            self.viewset.exclude_by_tag(mock_request)

    def test_exclude_by_tag_with_pagination(self):
        mock_request = mock.MagicMock()
        mock_request.query_params = {"key": "test_tag"}
        mock_request.data = {}

        page = [self.mock_object]
        self.viewset.paginate_queryset.return_value = page

        paginated_response = Response({"results": {"data": "serialized_data"}})
        self.viewset.get_paginated_response = mock.MagicMock(
            return_value=paginated_response
        )

        self.viewset.taggable_service.exclude_by_tag.return_value = (
            self.mock_filtered_queryset
        )

        response = self.viewset.exclude_by_tag(mock_request)

        self.viewset.paginate_queryset.assert_called_once_with(
            self.mock_filtered_queryset
        )

        self.viewset.get_serializer.assert_called_once_with(page, many=True)

        self.viewset.get_paginated_response.assert_called_once_with(
            {"data": "serialized_data"}
        )

        self.assertEqual(response, paginated_response)



SECRET_KEY = "SECRET"
BROKER_URL = "memory://"

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "flexi_tag.tests",
    "flexi_tag",
]

DEFAULT_AUTO_FIELD = "django.db.models.AutoField"

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.postgresql",
        "NAME": "postgres",
        "USER": "postgres",
        "PASSWORD": "postgres",
        "HOST": "localhost",
        "PORT": "5432",
    }
}

REST_FRAMEWORK = {}
TIME_ZONE = "UTC"



from rest_framework import status
from rest_framework.decorators import action
from rest_framework.response import Response

from flexi_tag.exceptions import (
    ObjectIDsNotDefinedException,
    TagNotDefinedException,
    TagNotFoundException,
)
from flexi_tag.utils.service import TaggableService


class TaggableViewSetMixin(object):
    taggable_service = TaggableService()

    @action(detail=True, methods=["post"])
    def add_tag(self, request, pk=None):
        obj = self.get_object()  # noqa
        key = request.data.get("key", None)

        if not key:
            raise TagNotDefinedException()

        self.taggable_service.add_tag(
            instance=obj,
            key=key,
        )
        return Response(status=status.HTTP_200_OK)

    @action(detail=True, methods=["post"])
    def bulk_add_tag(self, request, pk=None):
        obj = self.get_object()  # noqa
        keys = request.data.get("keys", [])

        if not keys:
            raise TagNotDefinedException()

        self.taggable_service.bulk_add_tags(
            instance=obj,
            keys=keys,
        )
        return Response(status=status.HTTP_200_OK)

    @action(detail=False, methods=["post"])
    def bulk_add_tags(self, request, pk=None):
        objs = request.data.get("objects", [])
        keys = request.data.get("keys", [])

        if not keys:
            raise TagNotDefinedException()
        if not objs:
            raise ObjectIDsNotDefinedException()

        objs = self.get_queryset().filter(id__in=objs)  # noqa

        self.taggable_service.bulk_add_tags_with_many_instances(
            instances=objs,
            keys=keys,
        )
        return Response(status=status.HTTP_200_OK)

    @action(detail=True, methods=["post"])
    def remove_tag(self, request, pk=None):
        obj = self.get_object()  # noqa
        key = request.data.get("key")

        if not key:
            raise TagNotFoundException()

        self.taggable_service.remove_tag(
            instance=obj,
            key=key,
        )
        return Response(status=status.HTTP_200_OK)

    @action(detail=True, methods=["post"])
    def bulk_remove_tags(self, request, pk=None):
        obj = self.get_object()  # noqa
        keys = request.data.get("keys", [])

        if not keys:
            raise TagNotDefinedException()

        self.taggable_service.bulk_remove_tags(
            instance=obj,
            keys=keys,
        )
        return Response(status=status.HTTP_200_OK)

    @action(detail=False, methods=["post"])
    def bulk_remove_tags_with_many_instances(self, request, pk=None):
        objs = request.data.get("objects", [])
        keys = request.data.get("keys", [])

        if not keys:
            raise TagNotDefinedException()
        if not objs:
            raise ObjectIDsNotDefinedException()

        objs = self.get_queryset().filter(id__in=objs)  # noqa

        self.taggable_service.bulk_remove_tags_with_many_instances(
            instances=objs,
            keys=keys,
        )
        return Response(status=status.HTTP_200_OK)

    @action(detail=True, methods=["get"])
    def get_tags(self, request, pk=None):
        obj = self.get_object()
        tag_instance = self.taggable_service.get_tags(instance=obj)
        if not tag_instance:
            return Response(status=status.HTTP_404_NOT_FOUND)
        return Response(tag_instance, status=status.HTTP_200_OK)

    @action(detail=False, methods=["get"])
    def filter_by_tag(self, request):
        key = request.query_params.get("key") or request.data.get("key")

        if not key:
            raise TagNotDefinedException()

        queryset = self.taggable_service.filter_by_tag(
            queryset=self.get_queryset(),  # noqa
            key=key,
        )

        page = self.paginate_queryset(queryset)  # noqa
        if page is not None:
            serializer = self.get_serializer(page, many=True)  # noqa
            return self.get_paginated_response(serializer.data)  # noqa

        serializer = self.get_serializer(queryset, many=True)  # noqa
        return Response(serializer.data)

    @action(detail=False, methods=["get"])
    def exclude_by_tag(self, request):
        key = request.query_params.get("key") or request.data.get("key")

        if not key:
            raise TagNotDefinedException()

        queryset = self.taggable_service.exclude_by_tag(
            queryset=self.get_queryset(),  # noqa
            key=key,
        )

        page = self.paginate_queryset(queryset)  # noqa
        if page is not None:
            serializer = self.get_serializer(page, many=True)  # noqa
            return self.get_paginated_response(serializer.data)  # noqa

        serializer = self.get_serializer(queryset, many=True)  # noqa
        return Response(serializer.data)



"""
Compatibility module for Django version differences.
"""
import django

# JSONField location changed in Django 3.1
if django.VERSION >= (3, 1):
    from django.db.models import JSONField
else:
    from django.contrib.postgres.fields import JSONField



from django.apps import apps
from django.db import models


class TaggableManager(models.Manager):
    @staticmethod
    def __validate_tag_model_class(app_label, tag_model_name):
        try:
            return apps.get_model(app_label, tag_model_name)
        except LookupError:
            raise ValueError(
                "Tag model for {app_label}.{model_name} not found. Did you run 'python manage.py generate_tag_models', generate migration file and apply it?".format(
                    app_label=app_label,
                    model_name=tag_model_name,
                )
            )

    def __get_tag_model_class_name(self, model_class):
        model_name = model_class.__name__
        app_label = model_class._meta.app_label  # noqa
        tag_model_name = "{}Tag".format(model_name)
        self.__validate_tag_model_class(
            app_label=app_label,
            tag_model_name=tag_model_name,
        )

        return tag_model_name

    def filter_tag(self, key):
        model_class = self.model
        tag_model_class_name = self.__get_tag_model_class_name(model_class)

        query_filter = {"{}__tags__contains".format(tag_model_class_name): [key]}
        return self.get_queryset().filter(**query_filter)

    def exclude_tag(self, key):
        model_class = self.model
        tag_model_class_name = self.__get_tag_model_class_name(model_class)

        query_filter = {"{}__tags__contains".format(tag_model_class_name): [key]}
        return self.get_queryset().exclude(**query_filter)

    def with_tags(self):
        model_class = self.model
        tag_model_class_name = self.__get_tag_model_class_name(model_class)

        return self.get_queryset().prefetch_related(tag_model_class_name)



try:
    from StringIO import StringIO  # Python 2
except ImportError:
    from io import StringIO  # Python 3
import os
import shutil
import tempfile

try:
    from unittest import mock
except ImportError:
    import mock

from django.core.management import call_command
from django.db import models
from django.template import Context, Engine, Template
from django.test import TestCase

from flexi_tag.management.commands.generate_tag_models import Command
from flexi_tag.utils.models import FlexiTagMixin


class TestModel(FlexiTagMixin):
    class Meta:
        app_label = "tests"


class SimpleTestModel(FlexiTagMixin):
    name = models.CharField(max_length=100)

    class Meta:
        app_label = "tests"


class NestedTestModel(FlexiTagMixin):
    title = models.CharField(max_length=100)
    parent = models.ForeignKey("self", null=True, blank=True, on_delete=models.CASCADE)

    class Meta:
        app_label = "tests"


try:
    import builtins  # noqa: F401

    open_path = "builtins.open"
except ImportError:
    import __builtin__  # noqa: F401

    open_path = "__builtin__.open"


class GenerateTagModelsCommandTestCase(TestCase):
    def setUp(self):
        self.test_dir = tempfile.mkdtemp()

        self.model_template_string = """
# This file is auto-generated. Do not edit manually.
# Generated by the generate_tag_models command.

from django.db import models
from flexi_tag.utils.compat import JSONField
from django.contrib.postgres.indexes import GinIndex


class {{ model_name }}Tag(models.Model):
    instance = models.OneToOneField(
        "{{ app_label }}.{{ model_name }}",
        on_delete=models.CASCADE,
        primary_key=True,
    )
    tags = JSONField(default=list)

    class Meta:
        app_label = "{{ app_label }}"
        db_table = "{{ app_label }}_{{ model_lower_name }}_tag"
        indexes = [GinIndex(fields=["tags"])]

    def __str__(self):
        return "Tags for {}".format(self.instance)"
"""

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    @mock.patch(
        "flexi_tag.management.commands.generate_tag_models.apps.get_app_configs"
    )
    @mock.patch("flexi_tag.management.commands.generate_tag_models.import_module")
    @mock.patch("flexi_tag.management.commands.generate_tag_models.inspect.getfile")
    @mock.patch("flexi_tag.management.commands.generate_tag_models.Template")
    def test_command_execution(
        self, mock_template, mock_getfile, mock_import_module, mock_get_app_configs
    ):
        mock_model = mock.MagicMock()
        mock_model.__name__ = "MyModel"
        mock_model._meta.app_label = "myapp"
        mock_model.__module__ = "myapp.models"

        mock_model.__bases__ = (FlexiTagMixin,)

        mock_app_config = mock.MagicMock()
        mock_app_config.get_models.return_value = [mock_model]
        mock_get_app_configs.return_value = [mock_app_config]

        mock_module = mock.MagicMock()
        mock_import_module.return_value = mock_module

        test_file_path = os.path.join(self.test_dir, "models.py")
        mock_getfile.return_value = test_file_path

        mock_template_instance = mock.MagicMock()
        mock_template_instance.render.return_value = "# Generated model content"
        mock_template.return_value = mock_template_instance

        out = StringIO()

        call_command("generate_tag_models", stdout=out)

        self.assertIn("Successfully generated all tag models", out.getvalue())

    @mock.patch(
        "flexi_tag.management.commands.generate_tag_models.apps.get_app_configs"
    )
    @mock.patch("flexi_tag.management.commands.generate_tag_models.import_module")
    @mock.patch("flexi_tag.management.commands.generate_tag_models.inspect.getfile")
    @mock.patch("flexi_tag.management.commands.generate_tag_models.Template")
    @mock.patch(open_path, new_callable=mock.mock_open)
    def test_dry_run_option(
        self,
        mock_open,
        mock_template,
        mock_getfile,
        mock_import_module,
        mock_get_app_configs,
    ):
        mock_model = mock.MagicMock()
        mock_model.__name__ = "MyModel"
        mock_model._meta.app_label = "myapp"
        mock_model.__module__ = "myapp.models"

        mock_model.__bases__ = (FlexiTagMixin,)

        mock_app_config = mock.MagicMock()
        mock_app_config.get_models.return_value = [mock_model]
        mock_get_app_configs.return_value = [mock_app_config]

        mock_module = mock.MagicMock()
        mock_import_module.return_value = mock_module

        test_file_path = os.path.join(self.test_dir, "models.py")
        mock_getfile.return_value = test_file_path

        mock_template_instance = mock.MagicMock()
        mock_template_instance.render.return_value = "# Generated model content"
        mock_template.return_value = mock_template_instance

        out = StringIO()
        call_command("generate_tag_models", "--dry-run", stdout=out)

        self.assertIn("Dry run completed", out.getvalue())

        self.assertFalse(mock_open.called)

    @mock.patch("django.template.Engine.get_default")
    def test_template_content(self, mock_get_default):
        engine = Engine()
        mock_get_default.return_value = engine

        template = Template(self.model_template_string)

        context = Context(
            {
                "model_name": "TestModel",
                "app_label": "testapp",
                "model_lower_name": "testmodel",
            }
        )

        rendered_template = template.render(context)

        self.assertIn("class TestModelTag(models.Model):", rendered_template)
        self.assertIn('app_label = "testapp"', rendered_template)
        self.assertIn('db_table = "testapp_testmodel_tag"', rendered_template)

    @mock.patch(
        "flexi_tag.management.commands.generate_tag_models.apps.get_app_configs"
    )
    @mock.patch("flexi_tag.management.commands.generate_tag_models.Template")
    def test_command_with_no_flexi_tag_models(
        self, mock_template, mock_get_app_configs
    ):
        mock_template_instance = mock.MagicMock()
        mock_template_instance.render.return_value = "# Generated model content"
        mock_template.return_value = mock_template_instance

        out = StringIO()

        mock_model = mock.MagicMock()
        mock_model.__name__ = "RegularModel"
        mock_model.__bases__ = (object,)

        mock_app_config = mock.MagicMock()
        mock_app_config.get_models.return_value = [mock_model]
        mock_get_app_configs.return_value = [mock_app_config]

        call_command("generate_tag_models", stdout=out)

        self.assertIn("Successfully generated all tag models", out.getvalue())


class CustomTagGenerationTestCase(TestCase):
    def setUp(self):
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    @mock.patch("flexi_tag.management.commands.generate_tag_models.Template")
    @mock.patch(
        "flexi_tag.management.commands.generate_tag_models.apps.get_app_configs"
    )
    @mock.patch("flexi_tag.management.commands.generate_tag_models.import_module")
    @mock.patch("flexi_tag.management.commands.generate_tag_models.inspect.getfile")
    @mock.patch("flexi_tag.management.commands.generate_tag_models.open", create=True)
    def test_multiple_models_generation(
        self,
        mock_open,
        mock_getfile,
        mock_import_module,
        mock_get_app_configs,
        mock_template_class,
    ):
        mock_template = mock.MagicMock()
        mock_template.render.return_value = "# Mocked template content"
        mock_template_class.return_value = mock_template

        mock_model1 = mock.MagicMock()
        mock_model1.__name__ = "Model1"
        mock_model1._meta.app_label = "testapp"
        mock_model1.__module__ = "testapp.models"
        mock_model1.__bases__ = (FlexiTagMixin,)

        mock_model2 = mock.MagicMock()
        mock_model2.__name__ = "Model2"
        mock_model2._meta.app_label = "testapp"
        mock_model2.__module__ = "testapp.models"
        mock_model2.__bases__ = (FlexiTagMixin,)

        mock_app_config = mock.MagicMock()
        mock_app_config.get_models.return_value = [mock_model1, mock_model2]
        mock_get_app_configs.return_value = [mock_app_config]

        mock_module = mock.MagicMock()
        mock_import_module.return_value = mock_module

        test_file_path = os.path.join(self.test_dir, "models.py")
        mock_getfile.return_value = test_file_path

        mock_file = mock.MagicMock()
        mock_open.return_value.__enter__.return_value = mock_file

        out = StringIO()

        call_command("generate_tag_models", stdout=out)

        self.assertEqual(mock_open.call_count, 2)
        self.assertEqual(mock_file.write.call_count, 2)
        self.assertIn("Successfully generated all tag models", out.getvalue())

    @mock.patch("flexi_tag.management.commands.generate_tag_models.Template")
    @mock.patch(
        "flexi_tag.management.commands.generate_tag_models.apps.get_app_configs"
    )
    @mock.patch("flexi_tag.management.commands.generate_tag_models.import_module")
    @mock.patch("flexi_tag.management.commands.generate_tag_models.inspect.getfile")
    @mock.patch("flexi_tag.management.commands.generate_tag_models.open", create=True)
    def test_model_with_special_name(
        self,
        mock_open,
        mock_getfile,
        mock_import_module,
        mock_get_app_configs,
        mock_template_class,
    ):
        mock_template = mock.MagicMock()
        mock_template.render.return_value = "# Mocked template content"
        mock_template_class.return_value = mock_template

        mock_model = mock.MagicMock()
        mock_model.__name__ = "Special_Model_Name-With.Characters$"
        mock_model._meta.app_label = "testapp"
        mock_model.__module__ = "testapp.models"
        mock_model.__bases__ = (FlexiTagMixin,)

        mock_app_config = mock.MagicMock()
        mock_app_config.get_models.return_value = [mock_model]
        mock_get_app_configs.return_value = [mock_app_config]

        mock_module = mock.MagicMock()
        mock_import_module.return_value = mock_module

        test_file_path = os.path.join(self.test_dir, "models.py")
        mock_getfile.return_value = test_file_path

        mock_file = mock.MagicMock()
        mock_open.return_value.__enter__.return_value = mock_file

        out = StringIO()

        call_command("generate_tag_models", stdout=out)

        self.assertEqual(mock_open.call_count, 1)
        self.assertEqual(mock_file.write.call_count, 1)
        self.assertEqual(mock_template.render.call_count, 1)

        context = mock_template.render.call_args[0][0]
        self.assertEqual(context["model_name"], "Special_Model_Name-With.Characters$")
        self.assertEqual(
            context["model_lower_name"], "special_model_name-with.characters$"
        )
        self.assertIn("Successfully generated all tag models", out.getvalue())

    @mock.patch("flexi_tag.management.commands.generate_tag_models.Template")
    @mock.patch(
        "flexi_tag.management.commands.generate_tag_models.apps.get_app_configs"
    )
    @mock.patch("flexi_tag.management.commands.generate_tag_models.import_module")
    @mock.patch("flexi_tag.management.commands.generate_tag_models.inspect.getfile")
    @mock.patch("flexi_tag.management.commands.generate_tag_models.open", create=True)
    def test_command_with_non_django_class(
        self,
        mock_open,
        mock_getfile,
        mock_import_module,
        mock_get_app_configs,
        mock_template_class,
    ):
        mock_template = mock.MagicMock()
        mock_template.render.return_value = "# Mocked template content"
        mock_template_class.return_value = mock_template

        mock_model = mock.MagicMock()
        mock_model.__name__ = "NonModelClass"
        mock_model.__bases__ = (FlexiTagMixin,)
        mock_model._meta = mock.MagicMock()
        mock_model._meta.app_label = "testapp"
        mock_model.__module__ = "testapp.models"

        mock_module = mock.MagicMock()
        mock_import_module.return_value = mock_module

        test_file_path = os.path.join(self.test_dir, "models.py")
        mock_getfile.return_value = test_file_path

        mock_file = mock.MagicMock()
        mock_open.return_value.__enter__.return_value = mock_file

        mock_app_config = mock.MagicMock()
        mock_app_config.get_models.return_value = [mock_model]
        mock_get_app_configs.return_value = [mock_app_config]

        out = StringIO()

        call_command("generate_tag_models", stdout=out)

        self.assertIn("Successfully generated all tag models", out.getvalue())

    @mock.patch("sys.exit")
    def test_command_help(self, mock_exit):
        command = Command()
        command.stdout = StringIO()

        help_text = command.help

        self.assertIn("Generate tag models", help_text)
        self.assertTrue(callable(command.add_arguments))

        parser = mock.MagicMock()
        command.add_arguments(parser)
        self.assertTrue(parser.add_argument.called)
        args, _ = parser.add_argument.call_args
        self.assertEqual(args[0], "--dry-run")

        mock_exit.assert_not_called()

    @mock.patch("flexi_tag.management.commands.generate_tag_models.Template")
    @mock.patch(
        "flexi_tag.management.commands.generate_tag_models.apps.get_app_configs"
    )
    @mock.patch("flexi_tag.management.commands.generate_tag_models.import_module")
    @mock.patch("flexi_tag.management.commands.generate_tag_models.inspect.getfile")
    def test_dry_run_option(
        self,
        mock_getfile,
        mock_import_module,
        mock_get_app_configs,
        mock_template_class,
    ):
        mock_template = mock.MagicMock()
        mock_template.render.return_value = "# Mocked template content"
        mock_template_class.return_value = mock_template

        mock_model = mock.MagicMock()
        mock_model.__name__ = "TestModel"
        mock_model._meta.app_label = "testapp"
        mock_model.__module__ = "testapp.models"
        mock_model.__bases__ = (FlexiTagMixin,)

        mock_app_config = mock.MagicMock()
        mock_app_config.get_models.return_value = [mock_model]
        mock_get_app_configs.return_value = [mock_app_config]

        mock_module = mock.MagicMock()
        mock_import_module.return_value = mock_module

        test_file_path = os.path.join(self.test_dir, "models.py")
        mock_getfile.return_value = test_file_path

        out = StringIO()

        call_command("generate_tag_models", "--dry-run", stdout=out)

        output = out.getvalue()
        self.assertIn("Would create file:", output)
        self.assertIn("Dry run completed. No files were created.", output)
        self.assertNotIn("Successfully generated all tag models", output)



from django.apps import apps

from flexi_tag.exceptions import TagNotFoundException, TagValidationException


class TaggableService:
    @staticmethod
    def __get_tag_model_class(instance):
        model_name = instance.__class__.__name__
        app_label = instance._meta.app_label  # noqa

        tag_model = "{}Tag".format(model_name)
        try:
            return apps.get_model(app_label, tag_model)
        except LookupError:
            raise ValueError(
                "Tag model for {app_label}.{model_name} not found. Did you run 'python manage.py generate_tag_models', generate migration file and apply it?".format(
                    app_label=app_label,
                    model_name=model_name,
                )
            )

    @staticmethod
    def __validate_tag_key(key):
        if not isinstance(key, str):
            raise ValueError("Tag key must be a string.")
        return True

    def add_tag(self, instance, key):
        self.__validate_tag_key(key)
        tag_model = TaggableService.__get_tag_model_class(instance)
        tag_instance, created = tag_model.objects.get_or_create(instance=instance)

        tags = tag_instance.tags
        if not created and key in tags:
            raise TagValidationException(name=key)

        tags.append(key)

        tag_instance.tags = tags
        tag_instance.save(update_fields=["tags"])

        return instance

    def bulk_add_tags(self, instance, keys):
        tag_model = TaggableService.__get_tag_model_class(instance)
        tag_instance, created = tag_model.objects.get_or_create(instance=instance)

        tags = tag_instance.tags
        if not created:
            for key in keys:
                self.__validate_tag_key(key)
                if key in tags:
                    raise TagValidationException(name=key)

        tags.extend(keys)
        tag_instance.tags = tags
        tag_instance.save(update_fields=["tags"])

        return instance

    def bulk_add_tags_with_many_instances(self, instances, keys):
        tag_model = TaggableService.__get_tag_model_class(instances[0])
        tags = []

        for instance in instances.iterator():
            tag_instance, created = tag_model.objects.get_or_create(instance=instance)
            if not created:
                for key in keys:
                    self.__validate_tag_key(key)
                    if key in tag_instance.tags:
                        raise TagValidationException(name=key)

            tags.extend(keys)
            tag_instance.tags = tags
            tag_instance.save(update_fields=["tags"])

        return instances

    def remove_tag(self, instance, key):
        tag_model = self.__get_tag_model_class(instance)

        tag_instance = tag_model.objects.filter(instance=instance).first()
        if not tag_instance:
            raise TagNotFoundException(name=key)

        if key in tag_instance.tags:
            tag_instance.tags.remove(key)
            tag_instance.save(update_fields=["tags"])
        else:
            raise TagValidationException(name=key)

        return tag_instance

    def bulk_remove_tags(self, instance, keys):
        tag_model = self.__get_tag_model_class(instance)

        tag_instance = tag_model.objects.filter(instance=instance).first()
        if not tag_instance:
            raise TagNotFoundException(name=keys)

        for key in keys:
            if key in tag_instance.tags:
                tag_instance.tags.remove(key)
            else:
                raise TagValidationException(name=key)

        tag_instance.save(update_fields=["tags"])

        return tag_instance

    def bulk_remove_tags_with_many_instances(
        self,
        instances,
        keys,
    ):
        tag_model = self.__get_tag_model_class(instances[0])

        for instance in instances.iterator():
            tag_instance = tag_model.objects.filter(instance=instance).first()
            if not tag_instance:
                raise TagNotFoundException(name=keys)

            for key in keys:
                if key in tag_instance.tags:
                    tag_instance.tags.remove(key)
                else:
                    raise TagValidationException(name=key)

            tag_instance.save(update_fields=["tags"])

        return instances

    def get_tags(self, instance):
        tag_model = self.__get_tag_model_class(instance)
        tag_instance = tag_model.objects.get(instance=instance)
        if not tag_instance:
            return []

        return tag_instance.tags

    @staticmethod
    def filter_by_tag(queryset, key):
        return queryset.filter_tag(key)

    @staticmethod
    def exclude_by_tag(queryset, key):
        return queryset.exclude_tag(key)

    @staticmethod
    def get_with_tags(queryset):
        return queryset.with_tags()



from django.db import models

from flexi_tag.utils.model_managers import TaggableManager


class FlexiTagMixin(models.Model):
    """
    Mixin for tag model generation.
    This mixin doesn't add any fields to the model itself.
    It's used by the generate_tag_models command to identify models that should have
    a corresponding tag model generated.

    The tag model will be generated as:

    class YourModelTag(models.Model):
        instance = models.OneToOneField(
            "app_label.YourModel",
            on_delete=models.CASCADE,
            primary_key=True,
        )
        tags = JSONField(default=dict)  # Uses the appropriate JSONField for your Django version

        class Meta:
            indexes = [GinIndex(fields=["tags"])]

    Usage:
        from flexi_tag.utils.models import FlexiTagMixin

        class YourModel(FlexiTagMixin):
            # your model fields here
            pass

    After defining your models, run:
        python manage.py generate_tag_models

    This will create flexi_generated_model.py files in the same directories as your models.

    To use the tags functionality, use the FlexiTagService:
        from flexi_tag.utils.service import FlexiTagService

        # Add a tag
        FlexiTagService.add_tag(instance, 'key')

        # Get all tags
        tags_dict = FlexiTagService.get_tags(instance)

        # Remove a tag
        FlexiTagService.remove_tag(instance, 'key')
    """

    objects = TaggableManager()

    class Meta:
        abstract = True
